@@ -1,25 +1,33 @@
 import base64
 import functools
+import glob
 import io
+import json
 import logging
-import multiprocessing
 import os
 import platform
 import random
+import re
 import shutil
 import time
+from typing import List
+from contextlib import contextmanager
 
-from PIL import Image
 from botcity.base import BaseBot, State
 from botcity.base.utils import only_if_element
 from bs4 import BeautifulSoup
+from PIL import Image
 from selenium.common.exceptions import InvalidSessionIdException
 from selenium.webdriver.common.action_chains import ActionChains
+from selenium.common.exceptions import StaleElementReferenceException
+from selenium.webdriver.common.by import By
 from selenium.webdriver.common.keys import Keys
+from selenium.webdriver.remote.webelement import WebElement
+from selenium.webdriver.support.wait import WebDriverWait, TimeoutException, NoSuchElementException
+from selenium.webdriver.support import expected_conditions as EC
 
 from . import config, cv2find
-from .browsers import Browser, BROWSER_CONFIGS
-
+from .browsers import BROWSER_CONFIGS, Browser, PageLoadStrategy
 
 try:
     from botcity.maestro import BotMaestroSDK
@@ -33,6 +41,7 @@ logger = logging.getLogger(__name__)
 
 class WebBot(BaseBot):
     KEYS = Keys
+    DEFAULT_DIMENSIONS = (1600, 900)
     """
     Base class for Web Bots.
     Users must implement the `action` method in their classes.
@@ -49,22 +58,24 @@ class WebBot(BaseBot):
 
         self._browser = Browser.CHROME
         self._options = None
+        self._capabilities = None
         self._driver_path = None
 
         self._driver = None
         self._headless = headless
+        self._page_load_strategy = PageLoadStrategy.NORMAL
 
         self._clipboard = ""
 
         # Stub mouse coordinates
+        self._html_elem = None
         self._x = 0
         self._y = 0
 
         # State for Key modifiers
         self._shift_hold = False
 
-        self._dimensions = (1600, 900)
-        self._download_folder_path = None  # Defaults to ~/Desktop
+        self._download_folder_path = os.getcwd()
 
     @property
     def driver(self):
@@ -90,6 +101,7 @@ class WebBot(BaseBot):
             driver_path (str): The full path to the proper webdriver path used for the selected browser.
                 If set to None, the code will look into the PATH for the proper file when starting the browser.
         """
+        driver_path = os.path.abspath(os.path.expanduser(os.path.expandvars(driver_path)))
         if driver_path and not os.path.isfile(driver_path):
             raise ValueError("Invalid driver_path. The file does not exist.")
         self._driver_path = driver_path
@@ -134,6 +146,26 @@ class WebBot(BaseBot):
         """
         self._options = options
 
+    @property
+    def capabilities(self):
+        """
+        The capabilities to be passed down to the WebDriver when starting the browser.
+
+        Returns:
+            capabilities (Dict): The browser specific capabilities to be used.
+        """
+        return self._capabilities
+
+    @capabilities.setter
+    def capabilities(self, capabilities):
+        """
+        The capabilities to be passed down to the WebDriver when starting the browser.
+
+        Args:
+            capabilities (Dict): The browser specific capabilities to be used.
+        """
+        self._capabilities = capabilities
+
     @property
     def download_folder_path(self):
         return self._download_folder_path
@@ -171,6 +203,28 @@ class WebBot(BaseBot):
             logger.warning("Browser is running. Invoke stop_browser and start browser for changes to take effect.")
         self._headless = headless
 
+    @property
+    def page_load_strategy(self) -> PageLoadStrategy:
+        """
+        The page load strategy to be used.
+
+        Returns:
+            page_load_strategy (PageLoadStrategy): The page load strategy to be used.
+        """
+        return self._page_load_strategy
+
+    @page_load_strategy.setter
+    def page_load_strategy(self, page_load_strategy: PageLoadStrategy):
+        """
+        The page load strategy to be used.
+
+        Args:
+            page_load_strategy (PageLoadStrategy): The page load strategy to be used.
+        """
+        if self._driver:
+            logger.warning("Browser is running. Invoke stop_browser and start browser for changes to take effect.")
+        self._page_load_strategy = page_load_strategy
+
     def start_browser(self):
         """
         Starts the selected browser.
@@ -189,22 +243,34 @@ class WebBot(BaseBot):
         driver_class = BROWSER_CONFIGS.get(self.browser).get("class")
         # Specific default options method for a given browser
         func_def_options = BROWSER_CONFIGS.get(self.browser).get("options")
+        # Specific capabilities method for a given browser
+        func_def_capabilities = BROWSER_CONFIGS.get(self.browser).get("capabilities")
 
-        opt = self.options or func_def_options(self.headless, self._download_folder_path, None)
+        opt = self.options or func_def_options(
+            self.headless, self._download_folder_path, None, self.page_load_strategy
+        )
+        cap = self.capabilities or func_def_capabilities()
         self.options = opt
+        self.capabilities = cap
         driver_path = self.driver_path or check_driver()
         self.driver_path = driver_path
 
-        self._driver = driver_class(options=opt, executable_path=driver_path)
+        self._driver = driver_class(options=opt, desired_capabilities=cap, executable_path=driver_path)
         self.set_screen_resolution()
 
     def stop_browser(self):
         """
         Stops the Chrome browser and clean up the User Data Directory.
+
+        Warning:
+            After invoking this method, you will need to reassign your custom options and capabilities.
         """
         if not self._driver:
             return
+        self._driver.close()
         self._driver.quit()
+        self.options = None
+        self.capabilities = None
         self._driver = None
 
     def set_screen_resolution(self, width=None, height=None):
@@ -215,21 +281,45 @@ class WebBot(BaseBot):
             width (int): The desired width.
             height (int): The desired height.
         """
-        width = width or self._dimensions[0]
-        height = height or self._dimensions[1]
+        dimensions = (width or self.DEFAULT_DIMENSIONS[0], height or self.DEFAULT_DIMENSIONS[1])
 
         if self.headless:
             # When running headless the window size is the viewport size
-            window_size = (width, height)
+            window_size = dimensions
         else:
             # When running non-headless we need to account for the borders and etc
             # So the size must be bigger to have the same viewport size as before
             window_size = self._driver.execute_script("""
                 return [window.outerWidth - window.innerWidth + arguments[0],
                   window.outerHeight - window.innerHeight + arguments[1]];
-                """, width, height)
+                """, *dimensions)
         self._driver.set_window_size(*window_size)
 
+    def _webdriver_command(self, command, params=None, req_type="POST"):
+        """
+        Execute a webdriver command.
+
+        Args:
+            command (str): The command URL after the session part
+            params (dict): The payload to be serialized and sent to the webdriver. Defaults to None.
+            req_type (str, optional): The type of request to be made. Defaults to "POST".
+
+        Returns:
+            str: The value of the response
+        """
+        if not params:
+            params = {}
+
+        resource = f"/session/{self.driver.session_id}/{command}"
+        url = self.driver.command_executor._url + resource
+        body = json.dumps(params)
+        response = self.driver.command_executor._request(req_type, url, body)
+
+        if not response:
+            raise Exception(response.get('value'))
+
+        return response.get('value')
+
     ##########
     # Display
     ##########
@@ -248,13 +338,18 @@ class WebBot(BaseBot):
         if not region:
             region = (0, 0, 0, 0)
 
-        x = region[0] or 0
-        y = region[1] or 0
-        width = region[2] or self._dimensions[0]
-        height = region[3] or self._dimensions[1]
-        data = self._driver.get_screenshot_as_base64()
-        image_data = base64.b64decode(data)
-        img = Image.open(io.BytesIO(image_data))
+        x = region[0]
+        y = region[1]
+        width = region[2] or self._get_page_size()[0]
+        height = region[3] or self._get_page_size()[1]
+
+        try:
+            data = self._driver.get_screenshot_as_base64()
+            image_data = base64.b64decode(data)
+            img = Image.open(io.BytesIO(image_data))
+        except:  # noqa: E722
+            img = Image.new("RGB", (width, height))
+
         img = img.crop((x, y, x + width, y + height))
         return img
 
@@ -271,6 +366,21 @@ class WebBot(BaseBot):
         height = self._driver.get_window_size().get("height")
         return width, height
 
+    def _get_page_size(self):
+        """
+        Returns the browser current page size.
+
+        Returns:
+            width (int): The current page width.
+            height (int): The current page height.
+        """
+        if not self._driver:
+            return self.DEFAULT_DIMENSIONS
+
+        width = self.execute_javascript("return window.innerWidth")
+        height = self.execute_javascript("return window.innerHeight")
+        return width, height
+
     def add_image(self, label, path):
         """
         Add an image into the state image map.
@@ -327,7 +437,7 @@ class WebBot(BaseBot):
         def _to_dict(lbs, elems):
             return {k: v for k, v in zip(lbs, elems)}
 
-        screen_w, screen_h = self._dimensions
+        screen_w, screen_h = self._get_page_size()
         x = x or 0
         y = y or 0
         w = width or screen_w
@@ -336,7 +446,8 @@ class WebBot(BaseBot):
         region = (x, y, w, h)
 
         results = [None] * len(labels)
-        paths = [self.state.map_images[la] for la in labels]
+        paths = [self._search_image_file(la) for la in labels]
+        paths = [self._image_path_as_image(la) for la in paths]
 
         if threshold:
             # TODO: Figure out how we should do threshold
@@ -347,7 +458,6 @@ class WebBot(BaseBot):
             print('Warning: Ignoring best=False for now. It will be supported in the future.')
 
         start_time = time.time()
-        n_cpus = multiprocessing.cpu_count() - 1
 
         while True:
             elapsed_time = (time.time() - start_time) * 1000
@@ -357,8 +467,7 @@ class WebBot(BaseBot):
             haystack = self.screenshot()
             helper = functools.partial(self._find_multiple_helper, haystack, region, matching, grayscale)
 
-            with multiprocessing.Pool(processes=n_cpus) as pool:
-                results = pool.map(helper, paths)
+            results = [helper(p) for p in paths]
 
             results = [r for r in results]
             if None in results:
@@ -370,6 +479,10 @@ class WebBot(BaseBot):
         ele = cv2find.locate_all_opencv(
             needle, haystack, region=region, confidence=confidence, grayscale=grayscale
         )
+        try:
+            ele = next(ele)
+        except StopIteration:
+            ele = None
         return ele
 
     def find(self, label, x=None, y=None, width=None, height=None, *,
@@ -426,7 +539,7 @@ class WebBot(BaseBot):
             element (NamedTuple): The element coordinates. None if not found.
         """
         self.state.element = None
-        screen_w, screen_h = self._dimensions
+        screen_w, screen_h = self._get_page_size()
         x = x or 0
         y = y or 0
         w = width or screen_w
@@ -435,6 +548,7 @@ class WebBot(BaseBot):
         region = (x, y, w, h)
 
         element_path = self._search_image_file(label)
+        element_path = self._image_path_as_image(element_path)
 
         if threshold:
             # TODO: Figure out how we should do threshold
@@ -450,7 +564,7 @@ class WebBot(BaseBot):
             elapsed_time = (time.time() - start_time) * 1000
             if elapsed_time > waiting_time:
                 return None
-            haystack = self.get_screen_image(region=region)
+            haystack = self.get_screen_image()
             it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
                                            region=region, confidence=matching, grayscale=grayscale)
             try:
@@ -462,6 +576,115 @@ class WebBot(BaseBot):
                 self.state.element = ele
                 return ele
 
+    def set_current_element(self, element: cv2find.Box):
+        """
+        Changes the current screen element the bot will interact when using click(), move(), and similar methods.
+
+        This method is equivalent to self.state.element = element.
+
+        Args:
+            element (Box): A screen element from self.state.element or the find_all(as_list=True) method.
+        """
+        self.state.element = element
+
+    def find_all(self, label, x=None, y=None, width=None, height=None, *,
+                 threshold=None, matching=0.9, waiting_time=10000, grayscale=False, as_list: bool = False):
+        """
+        Find all elements defined by label on screen until a timeout happens.
+
+        Args:
+            label (str): The image identifier
+            x (int, optional): Search region start position x. Defaults to 0.
+            y (int, optional): Search region start position y. Defaults to 0.
+            width (int, optional): Search region width. Defaults to screen width.
+            height (int, optional): Search region height. Defaults to screen height.
+            threshold (int, optional): The threshold to be applied when doing grayscale search.
+                Defaults to None.
+            matching (float, optional): The matching index ranging from 0 to 1.
+                Defaults to 0.9.
+            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
+                Defaults to 10000ms (10s).
+            grayscale (bool, optional): Whether or not to convert to grayscale before searching.
+                Defaults to False.
+            as_list (bool, Optional): If True, returns a list of element coordinates instead of a generator.
+                Use set_active_element() to be able to interact with the found elements.
+                This parameter must be True if you intend to run multiple find_all() concurrently.
+                Defaults to False.
+
+        Returns:
+            elements (collections.Iterable[NamedTuple]): A generator with all element coordinates found.
+                None if not found.
+        """
+        def deduplicate(elems):
+            def find_same(item, items):
+                x_start = item.left
+                x_end = item.left + item.width
+                y_start = item.top
+                y_end = item.top + item.height
+                similars = []
+                for itm in items:
+                    if itm == item:
+                        continue
+                    if (itm.left >= x_start and itm.left < x_end)\
+                            and (itm.top >= y_start and itm.top < y_end):
+                        similars.append(itm)
+                        continue
+                return similars
+
+            index = 0
+            while True:
+                try:
+                    dups = find_same(elems[index], elems[index:])
+                    for d in dups:
+                        elems.remove(d)
+                    index += 1
+                except IndexError:
+                    break
+            return elems
+
+        self.state.element = None
+        screen_w, screen_h = self._get_page_size()
+        x = x or 0
+        y = y or 0
+        w = width or screen_w
+        h = height or screen_h
+
+        region = (x, y, w, h)
+
+        element_path = self._search_image_file(label)
+        element_path = self._image_path_as_image(element_path)
+
+        if threshold:
+            # TODO: Figure out how we should do threshold
+            print('Threshold not yet supported')
+
+        start_time = time.time()
+
+        while True:
+            elapsed_time = (time.time() - start_time) * 1000
+            if elapsed_time > waiting_time:
+                return None
+
+            haystack = self.get_screen_image()
+            it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
+                                           region=region, confidence=matching, grayscale=grayscale)
+
+            eles = [ele for ele in it]
+            if not eles:
+                continue
+            eles = deduplicate(list(eles))
+
+            # As List
+            if as_list:
+                return eles
+
+            # As Generator
+            for ele in eles:
+                if ele is not None:
+                    self.state.element = ele
+                    yield ele
+            break
+
     def find_text(self, label, x=None, y=None, width=None, height=None, *, threshold=None, matching=0.9,
                   waiting_time=10000, best=True):
         """
@@ -504,7 +727,7 @@ class WebBot(BaseBot):
         Returns:
             size (Tuple): The screen dimension (width and height) in pixels.
         """
-        return self._dimensions
+        return self._get_page_size()
 
     def screenshot(self, filepath=None, region=None):
         """
@@ -548,7 +771,7 @@ class WebBot(BaseBot):
         Returns:
             Image: The screenshot Image object
         """
-        screen_size = self._dimensions
+        screen_size = self._get_page_size()
         x = x or 0
         y = y or 0
         width = width or screen_size[0]
@@ -585,7 +808,7 @@ class WebBot(BaseBot):
             coords (Tuple): A tuple containing the x and y coordinates for the element.
         """
         self.state.element = None
-        screen_size = self._dimensions
+        screen_size = self._get_page_size()
         x = x or 0
         y = y or 0
         width = width or screen_size[0]
@@ -595,7 +818,12 @@ class WebBot(BaseBot):
         if not best:
             print('Warning: Ignoring best=False for now. It will be supported in the future.')
 
-        it = cv2find.locate_all_opencv(self.state.map_images[label], region=region, confidence=matching)
+        element_path = self._search_image_file(label)
+        element_path = self._image_path_as_image(element_path)
+        haystack = self.get_screen_image()
+
+        it = cv2find.locate_all_opencv(element_path, haystack_image=haystack,
+                                       region=region, confidence=matching)
         try:
             ele = next(it)
         except StopIteration:
@@ -685,6 +913,44 @@ class WebBot(BaseBot):
         """
         self.navigate_to(url)
 
+    def back(self):
+        """
+        Goes one step backward in the browser history.
+        """
+        self._driver.back()
+
+    def forward(self):
+        """
+        Goes one step forward in the browser history.
+        """
+        self._driver.forward()
+
+    def refresh(self):
+        """
+        Refreshes the current page.
+        """
+        self._driver.refresh()
+
+    @contextmanager
+    def wait_for_new_page(self, waiting_time=10000, activate=True):
+        """Context manager to wait for a new page to load and activate it.
+
+        Args:
+            waiting_time (int, optional): The maximum waiting time. Defaults to 10000.
+            activate (bool, optional): Whether or not to activate the new page. Defaults to True.
+
+        """
+        tabs = self.get_tabs()
+        yield
+        start_time = time.time()
+        while tabs == self.get_tabs():
+            elapsed_time = (time.time() - start_time) * 1000
+            if elapsed_time > waiting_time:
+                return None
+            time.sleep(0.1)
+        if activate:
+            self.activate_tab(self.get_tabs()[-1])
+
     def execute_javascript(self, code):
         """
         Execute the given javascript code.
@@ -733,12 +999,22 @@ class WebBot(BaseBot):
             return None
 
     def get_tabs(self):
+        """Get a list of tab handlers
+
+        Returns:
+            list: List of tab handlers
+        """
         try:
             return self._driver.window_handles
         except InvalidSessionIdException:
             return []
 
     def create_tab(self, url):
+        """Create a new tab and navigate to the given URL.
+
+        Args:
+            url (str): The desired URL.
+        """
         try:
             # Refactor this when Selenium 4 is released
             self.execute_javascript(f"window.open('{url}', '_blank');")
@@ -747,7 +1023,11 @@ class WebBot(BaseBot):
             self.navigate_to(url)
 
     def create_window(self, url):
+        """Creates a new window with the given URL.
 
+        Args:
+            url (str): The desired URL.
+        """
         try:
             # Refactor this when Selenium 4 is released
             self.execute_javascript(f"window.open('{url}', '_blank', 'location=0');")
@@ -756,6 +1036,8 @@ class WebBot(BaseBot):
             self.navigate_to(url)
 
     def close_page(self):
+        """Close the current active page (tab or window).
+        """
         try:
             self._driver.close()
 
@@ -767,13 +1049,277 @@ class WebBot(BaseBot):
             pass
 
     def activate_tab(self, handle):
+        """Activate a tab given by the handle.
+
+        Args:
+            handle (str): The tab or window handle.
+        """
         self._driver.switch_to.window(handle)
 
+    def print_pdf(self, path=None, print_options=None):
+        """Print the current page as a PDF file.
+
+        Args:
+            path (str, optional): The path for the file to be saved. Defaults to None.
+            print_options (dict, optional): Print options as defined at. Defaults to None.
+
+        Returns:
+            str: the saved file path
+        """
+        title = self.page_title() or "document"
+        title = re.sub("[\\\\|/:–]", "", title)
+        timeout = 60000
+        default_path = os.path.expanduser(os.path.join(self.download_folder_path, f"{title}.pdf"))
+
+        if self.browser in [Browser.CHROME, Browser.EDGE] and not self.headless:
+            pdf_current_count = self.get_file_count(file_extension=".pdf")
+            # Chrome still does not support headless webdriver print
+            # but Firefox does.
+            self.execute_javascript("window.print();")
+
+            # We need to wait for the file to be available in this case.
+            if self.page_title():
+                self.wait_for_file(default_path, timeout=timeout)
+            else:
+                # Waiting when the file don't have the page title in path
+                self.wait_for_new_file(file_extension=".pdf", current_count=pdf_current_count)
+
+            # Move the downloaded pdf file if the path is not None
+            if path:
+                last_downloaded_pdf = self.get_last_created_file(self.download_folder_path, ".pdf")
+                os.rename(last_downloaded_pdf, path)
+                return path
+            self.wait(2000)
+            return default_path
+
+        if print_options is None:
+            print_options = {
+                'landscape': False,
+                'displayHeaderFooter': False,
+                'printBackground': True,
+                'preferCSSPageSize': True,
+                'marginTop': 0,
+                'marginBottom': 0
+            }
+        data = self._webdriver_command("print", print_options)
+        bytes_file = base64.b64decode(data)
+        if not path:
+            path = default_path
+        with open(path, "wb") as f:
+            f.write(bytes_file)
+        return path
+
+    def wait_for_downloads(self, timeout: int = 120000):
+        """
+        Wait for all downloads to be finished.
+        Beware that this method replaces the current page with the downloads window.
+
+        Args:
+            timeout (int, optional): Timeout in millis. Defaults to 120000.
+        """
+        if self.browser in [Browser.CHROME, Browser.EDGE] and self.headless:
+            start_time = time.time()
+            while True:
+                elapsed_time = (time.time() - start_time) * 1000
+                if elapsed_time > timeout:
+                    return False
+                downloads_count = self.get_file_count(self.download_folder_path, ".crdownload")
+                if downloads_count == 0:
+                    return True
+                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
+
+        wait_method = BROWSER_CONFIGS.get(self.browser).get("wait_for_downloads")
+        # waits for all the files to be completed
+        WebDriverWait(self._driver, timeout/1000.0, 1).until(wait_method)
+
+    def find_elements(self, selector: str, by: By = By.CSS_SELECTOR,
+                      waiting_time=10000, ensure_visible: bool = True) -> List[WebElement]:
+        """Find elements using the specified selector with selector type specified by `by`.
+
+        Args:
+            selector (str): The selector string to be used.
+            by (str, optional): Selector type. Defaults to By.CSS_SELECTOR.
+                [See more](https://selenium-python.readthedocs.io/api.html#selenium.webdriver.common.by.By)
+            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
+                Defaults to 10000ms (10s).
+            ensure_visible (bool, optional): Whether to wait for the element to be visible. Defaults to True.
+
+        Returns:
+            List[WebElement]: List of elements found.
+
+        **Example:**
+        ```python
+        from botcity.web import By
+        ...
+        # Find element by ID
+        all_cells = self.find_elements("//td", By.XPATH)
+        ...
+        ```
+        """
+        if ensure_visible:
+            condition = EC.visibility_of_all_elements_located
+        else:
+            condition = EC.presence_of_all_elements_located
+
+        try:
+            elements = WebDriverWait(
+                self._driver, timeout=waiting_time / 1000.0
+            ).until(
+                condition((by, selector))
+            )
+            return elements
+        except (TimeoutException, NoSuchElementException) as ex:
+            print("Exception on find_elements", ex)
+            return None
+
+    def find_element(self, selector: str, by: str = By.CSS_SELECTOR, waiting_time=10000,
+                     ensure_visible: bool = False, ensure_clickable: bool = False) -> WebElement:
+        """Find an element using the specified selector with selector type specified by `by`.
+        If more than one element is found, the first instance is returned.
+
+        Args:
+            selector (str): The selector string to be used.
+            by (str, optional): Selector type. Defaults to By.CSS_SELECTOR.
+                [See more](https://selenium-python.readthedocs.io/api.html#selenium.webdriver.common.by.By)
+            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
+                Defaults to 10000ms (10s).
+            ensure_visible (bool, optional): Whether to wait for the element to be visible. Defaults to False.
+            ensure_clickable (bool, optional): Whether to wait for the element to be clickable. Defaults to False.
+                If True, `ensure_clickable` takes precedence over `ensure_visible`.
+
+        Returns:
+            WebElement: The element found.
+
+        **Example:**
+        ```python
+        from botcity.web import By
+        ...
+        # Find element by ID
+        elem = self.find_element("my_elem", By.ID)
+        # Find element by XPath
+        elem = self.find_element("//input[@type='submit']", By.XPATH)
+        ...
+        ```
+        """
+        condition = EC.visibility_of_element_located if ensure_visible else EC.presence_of_element_located
+        condition = EC.element_to_be_clickable if ensure_clickable else condition
+
+        try:
+            element = WebDriverWait(
+                self._driver, timeout=waiting_time/1000.0
+            ).until(
+                condition((by, selector))
+            )
+            return element
+        except (TimeoutException, NoSuchElementException):
+            return None
+
+    def scroll_element(self, element: WebElement, steps: int = 100, interval: float = 500,
+                       start: int = 0, end: int = None):
+        """Scrolls down an element by its scroll height or a given amount defined by `start` and `end`.
+
+        This is useful for scrolling down a page to load more content or
+        to scroll down a dynamically loaded element.
+
+        Args:
+            element (WebElement): The element to scroll.
+            steps (int, optional): Number of steps in which to conclude the scroll. Defaults to 100.
+            interval (float, optional): Time interval between each step. Defaults to 500ms.
+            start (int, optional): Start position. Defaults to 0.
+            end (int, optional): End position. Defaults to None.
+        """
+        ele_height = self.driver.execute_script(
+            "return arguments[0].scrollHeight;", element
+        )
+
+        start = max(0, start)
+        end = min(ele_height, end) if end is not None else ele_height
+
+        for i in range(start, end, steps):
+            self.driver.execute_script(
+                "arguments[0].scrollTo(0, arguments[1])", element, i)
+            self.sleep(interval/1000.0)
+
+    def wait_for_stale_element(self, element: WebElement, timeout: int = 10000):
+        """
+        Wait until the WebElement element becomes stale (outdated).
+
+        Args:
+            element (WebElement): The element to monitor for staleness.
+            timeout (int, optional): Timeout in millis. Defaults to 120000.
+        """
+        try:
+            WebDriverWait(self._driver, timeout=timeout/1000.0).until(EC.staleness_of(element))
+        except (TimeoutException, NoSuchElementException):
+            pass
+
+    def wait_for_element_visibility(self, element: WebElement, visible: bool = True, waiting_time=10000):
+        """Wait for the element to be visible or hidden.
+
+        Args:
+            element (WebElement): The element to wait for.
+            visible (bool, optional): Whether to wait for the element to be visible. Defaults to True.
+            waiting_time (int, optional): Maximum wait time (ms) to search for a hit.
+                Defaults to 10000ms (10s).
+        """
+        if visible:
+            wait_method = EC.visibility_of
+        else:
+            wait_method = EC.invisibility_of_element
+
+        WebDriverWait(self._driver, timeout=waiting_time/1000.0).until(wait_method(element))
+
+    def set_file_input_element(self, element: WebElement, filepath: str):
+        """Configure the filepath for upload in a file element.
+        Note: This method does not submit the form.
+
+        Args:
+            element (WebElement): The file upload element.
+            filepath (str): The path to the file to be uploaded.
+
+        **Example:**
+        ```python
+        ...
+        # Find element
+        elem = self.find_element("body > form > input[type=file]")
+        # Set the filepath
+        self.set_file_input_element(elem, "./test.txt")
+        ...
+        ```
+        """
+        fpath = os.path.abspath(os.path.expanduser(os.path.expandvars(filepath)))
+        element.send_keys(fpath)
+
+    def enter_iframe(self, iframe: WebElement):
+        """Switch the WebBot driver to the specified iframe.
+
+        Args:
+            iframe (WebElement): The desired iFrame.
+        """
+        self._driver.switch_to.frame(iframe)
+
+    def leave_iframe(self):
+        """Leave the iframe and switch the WebBot driver to the default content.
+        """
+        self._driver.switch_to.default_content()
+
+    def install_firefox_extension(self, extension):
+        """
+        Install an extension in the Firefox browser.
+        This will start the browser if it was not started yet.
+
+        Args:
+            extension (str): The path of the .xpi extension to be loaded.
+        """
+        if self.browser != Browser.FIREFOX:
+            raise ValueError("install_firefox_extension only works with Firefox.")
+        if not self._driver:
+            self.start_browser()
+        self._driver.install_addon(os.path.abspath(extension))
+
     #######
     # Mouse
     #######
-
-    @only_if_element
     def click_on(self, label):
         """
         Click on the element.
@@ -782,7 +1328,9 @@ class WebBot(BaseBot):
             label (str): The image identifier
         """
         x, y = self.get_element_coords_centered(label)
-        self.click(x, y)
+        if None in (x, y):
+            raise ValueError(f'Element not available. Cannot find {label}.')
+        self.click_at(x, y)
 
     @only_if_element
     def get_last_x(self):
@@ -812,7 +1360,20 @@ class WebBot(BaseBot):
             y (int): The Y coordinate
 
         """
-        # ActionChains(self._driver).move_by_offset(-self._x, -self._y).perform()
+        if self.browser == Browser.FIREFOX:
+            # Reset coordinates if the page has gone stale. Only required for Firefox
+            if self._html_elem is None:
+                self._html_elem = self._driver.find_element_by_tag_name('body')
+                self._x = 0
+                self._y = 0
+            else:
+                try:
+                    self._html_elem.is_enabled()
+                except StaleElementReferenceException:
+                    self._html_elem = self._driver.find_element_by_tag_name('body')
+                    self._x = 0
+                    self._y = 0
+
         mx = x - self._x
         my = y - self._y
         self._x = x
@@ -852,7 +1413,7 @@ class WebBot(BaseBot):
             wait_after (int, optional): Interval to wait after clicking on the element.
             clicks (int, optional): Number of times to click. Defaults to 1.
             interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
-            button (str, optional): One of 'left', 'right', 'middle'. Defaults to 'left'
+            button (str, optional): One of 'left', 'right'. Defaults to 'left'
         """
         x, y = self.state.center()
         self.click_at(x, y, clicks=clicks, button=button, interval_between_clicks=interval_between_clicks)
@@ -870,7 +1431,7 @@ class WebBot(BaseBot):
             wait_after (int, optional): Interval to wait after clicking on the element.
             clicks (int, optional): Number of times to click. Defaults to 1.
             interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
-            button (str, optional): One of 'left', 'right', 'middle'. Defaults to 'left'
+            button (str, optional): One of 'left', 'right'. Defaults to 'left'
         """
         x = self.state.x() + x
         y = self.state.y() + y
@@ -1022,7 +1583,7 @@ class WebBot(BaseBot):
             clicks (int, optional): Number of times to click. Defaults to 1.
             interval_between_clicks (int, optional): The interval between clicks in ms. Defaults to 0.
         """
-        self.click(clicks=clicks, button='right', interval=interval_between_clicks)
+        self.click(clicks=clicks, button='right', interval_between_clicks=interval_between_clicks)
         self.sleep(wait_after)
 
     def right_click_at(self, x, y):
@@ -1491,19 +2052,20 @@ class WebBot(BaseBot):
         """
         self.wait(interval)
 
-    def wait_for_file(self, path, timeout=10000):
+    def wait_for_file(self, path, timeout=60000):
         """
-        Invoke the system handler to open the given file.
+        Wait for a file to be available on disk.
 
         Args:
             path (str): The path for the file to be executed
             timeout (int, optional): Maximum wait time (ms) to search for a hit.
-                Defaults to 10000ms (10s).
+                Defaults to 60000ms (60s).
 
         Returns:
             status (bool): Whether or not the file was available before the timeout
 
         """
+        path = os.path.abspath(os.path.expanduser(os.path.expandvars(path)))
         start_time = time.time()
 
         while True:
@@ -1511,5 +2073,70 @@ class WebBot(BaseBot):
             if elapsed_time > timeout:
                 return False
             if os.path.isfile(path) and os.access(path, os.R_OK):
+                if self.browser == Browser.FIREFOX and os.path.getsize(path) == 0:
+                    # if file is empty, the download is not completed.
+                    continue
+                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
                 return True
             self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
+
+    def get_last_created_file(self, path=None, file_extension=""):
+        """Returns the last created file in a specific folder path.
+
+        Args:
+            path (str, optional): The path of the folder where the file is expected. Defaults to None.
+            file_extension (str, optional): The extension of the file to be searched for (e.g., .pdf, .txt).
+
+        Returns:
+            str: the path of the last created file
+        """
+        if not path:
+            path = self.download_folder_path
+
+        files_path = glob.glob(os.path.expanduser(os.path.join(path, f"*{file_extension}")))
+        last_created_file = max(files_path, key=os.path.getctime)
+        return last_created_file
+
+    def get_file_count(self, path=None, file_extension=""):
+        """Get the total number of files of the same type.
+
+        Args:
+            path (str, optional): The path of the folder where the files are saved.
+            file_extension (str, optional): The extension of the files to be searched for (e.g., .pdf, .txt).
+
+        Returns:
+            int: the number of files of the given type
+        """
+        if not path:
+            path = self.download_folder_path
+
+        files_path = glob.glob(os.path.expanduser(os.path.join(path, f"*{file_extension}")))
+        return len(files_path)
+
+    def wait_for_new_file(self, path=None, file_extension="", current_count=0, timeout=60000):
+        """
+        Wait for a new file to be available on disk without the file path.
+
+        Args:
+            path (str, optional): The path of the folder where the file is expected. Defaults to None.
+            file_extension (str, optional): The extension of the file to be searched for (e.g., .pdf, .txt).
+            current_count (int): The current number of files in the folder of the given type. Defaults to 0 files
+            timeout (int, optional): Maximum wait time (ms) to search for a hit.
+                Defaults to 60000ms (60s).
+
+        Returns:
+            str: the path of the last created file of the given type
+        """
+        if not path:
+            path = self.download_folder_path
+
+        start_time = time.time()
+        while True:
+            elapsed_time = (time.time() - start_time) * 1000
+            if elapsed_time > timeout:
+                return None
+            file_count = self.get_file_count(path, f"*{file_extension}")
+            if file_count == current_count + 1:
+                self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
+                return self.get_last_created_file(path, f"*{file_extension}")
+            self.sleep(config.DEFAULT_SLEEP_AFTER_ACTION)
