@@ -1,8 +1,11 @@
 import atexit
 import os
 import tempfile
+from typing import Dict
 
 from selenium import webdriver
+from selenium.webdriver import Firefox  # noqa: F401, F403
+from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
 from selenium.webdriver.firefox.options import Options as FirefoxOptions
 
 from ..util import cleanup_temp_dir
@@ -333,8 +336,27 @@ FIREFOX_MIMETYPES_TO_DOWNLOAD = ['application/vnd.hzn-3d-crossword', 'video/3gpp
                                  'application/vnd.zzazz.deck+xml']
 
 
-def default_options(headless=False, download_folder_path=None, user_data_dir=None):
+def default_options(headless=False, download_folder_path=None, user_data_dir=None,
+                    page_load_strategy="normal") -> FirefoxOptions:
+    """Retrieve the default options for this browser curated by BotCity.
+
+    Args:
+        headless (bool, optional): Whether or not to use the headless mode. Defaults to False.
+        download_folder_path (str, optional): The default path in which to save files.
+            If None, the current directory is used. Defaults to None.
+        user_data_dir ([type], optional): The directory to use as user profile.
+            If None, a new temporary directory is used. Defaults to None.
+        page_load_strategy (str, optional): The page load strategy to use.
+
+    Returns:
+        FirefoxOptions: The Firefox options.
+    """
     firefox_options = FirefoxOptions()
+    try:
+        page_load_strategy = page_load_strategy.value
+    except AttributeError:
+        page_load_strategy = page_load_strategy
+    firefox_options.page_load_strategy = page_load_strategy
     firefox_options.headless = headless
     if not user_data_dir:
         temp_dir = tempfile.TemporaryDirectory(prefix="botcity_")
@@ -345,14 +367,39 @@ def default_options(headless=False, download_folder_path=None, user_data_dir=Non
     firefox_profile.set_preference('browser.download.folderList', 2)
     firefox_profile.set_preference('browser.download.manager.showWhenStarting', False)
     if not download_folder_path:
-        download_folder_path = os.path.join(os.path.expanduser("~"), "Desktop")
+        download_folder_path = os.getcwd()
     firefox_profile.set_preference('browser.download.dir', download_folder_path)
     firefox_profile.set_preference('general.warnOnAboutConfig', False)
 
     mimetypes_to_download = ",".join(FIREFOX_MIMETYPES_TO_DOWNLOAD)
+    firefox_profile.set_preference("pdfjs.disabled", True)
+    firefox_profile.set_preference("plugin.disable_full_page_plugin_for_types", mimetypes_to_download)
     firefox_profile.set_preference('browser.helperApps.neverAsk.saveToDisk', mimetypes_to_download)
 
     firefox_profile.update_preferences()
     firefox_options.profile = firefox_profile
 
     return firefox_options
+
+
+def default_capabilities() -> Dict:
+    """Fetch the default capabilities for this browser.
+
+    Returns:
+        Dict: Dictionary with the default capabilities defined.
+    """
+    return DesiredCapabilities.FIREFOX.copy()
+
+
+def wait_for_downloads(driver):
+    """Wait for all downloads to finish.
+    *Important*: This method overwrites the current page with the downloads page.
+    """
+    if not driver.current_url.startswith("about:downloads"):
+        driver.get("about:downloads")
+
+    return driver.execute_script("""
+        var items = document.querySelector('richlistbox').itemChildren;
+        if (items.every(e => e.attributes["state"].value != 0))
+            return true;
+        """)
